"""
Cluster upstream sequences (UTR and leader) for each gene

For each gene, look at the sequences assigned to it. Take the upstream
sequences and compute a consensus for them. Only those assigned sequences are
taken that have a very low error rate for the V gene match.

Output a FASTA file that contains one consensus sequence for each gene.
"""
import logging
from collections import Counter

from .table import read_table
from .utils import iterative_consensus

logger = logging.getLogger(__name__)

# When computing a UTR consensus, ignore sequences that deviate by more than
# this factor from the median length.
UTR_MEDIAN_DEVIATION = 0.1


def add_arguments(parser):
	arg = parser.add_argument
	arg('--max-V-errors', '--max-error-percentage', '-e', dest='max_v_errors',
	    metavar='PERCENT', type=float, default=1,
		help='Allow PERCENT errors in V gene match. Default: %(default)s')
	arg('--max-FR1-errors', dest='max_fr1_errors', metavar='PERCENT', type=float, default=None,
	    help='Allow PERCENT errors in FR1 region.')
	arg('--max-CDR1-errors', dest='max_cdr1_errors', metavar='PERCENT', type=float, default=None,
	    help='Allow PERCENT errors in CDR1 region.')
	arg('--min-consensus-size', type=int, default=1, metavar='N',
	    help='Require at least N sequences for consensus. Default: %(default)s')
	arg('--consensus-threshold', '-t', metavar='PERCENT',
		type=float, default=75,
		help='Threshold for consensus computation. Default: %(default)s%%')
	arg('--no-ambiguous', '--no-N', default=False, action='store_true',
	    help='Discard consensus sequences with ambiguous bases')
	arg('--part', choices=['UTR', 'leader', 'UTR+leader'],
		default='UTR+leader', help='Which part of the sequence before the V '
		'gene match to analyze. Default: %(default)s')
	arg('--debug', default=False, action='store_true',
		help='Enable debugging output')
	arg('table', help='Table with parsed IgBLAST results (assigned.tab.gz or '
		'filtered.tab.gz)')


def main(args):
	if args.debug:
		logging.getLogger().setLevel(logging.DEBUG)
	table = read_table(args.table)
	n_genes = len(set(table['V_gene']))
	logger.info('%s rows read with %d unique gene names', len(table), n_genes)
	for name, column_name, arg in (
			('V%SHM', 'V_SHM', args.max_v_errors),
			('FR1%SHM', 'FR1_SHM', args.max_fr1_errors),
			('CDR1%SHM', 'CDR1_SHM', args.max_cdr1_errors)):
		if arg is not None:
			table = table[getattr(table, column_name) <= arg]
			logger.info('%s rows remain after discarding %s > %s%%',
				len(table), name, arg)
	table['UTR+leader'] = table['UTR'] + table['leader']
	table = table[table[args.part] != '']
	table['UTR_length'] = [ len(s) for s in table['UTR'] ]

	n_written = 0
	n_consensus_with_n = 0
	for name, group in table.groupby('V_gene'):
		assert len(group) != 0
		if len(group) < args.min_consensus_size:
			logger.info('Gene %s has too few assignments (%s), skipping.', name, len(group))
			continue

		counter = Counter(group['UTR_length'])
		logger.debug('Sequence length/count table: %s', ', '.join('{}: {}'.format(l,c) for l, c in counter.most_common()))

		if args.part == 'leader':
			sequences = list(group['leader'])
		else:
			# Since UTR lengths vary a lot, take all sequences that are at
			# least as long as the tenth longest sequence, and compute
			# consensus only from those.
			length_threshold = sorted(group['UTR_length'], reverse=True)[:10][-1]
			sequences = list(group[group['UTR_length'] >= length_threshold][args.part])

		if len(sequences) == 0:
			logger.info('Gene %s has %s assignments, but lengths are too different, skipping.', name, len(group))
			continue
		assert len(sequences) > 0
		if len(sequences) == 1:
			cons = sequences[0]
		else:
			# Keep only those sequences whose length is at least 90% of the longest one
			cons = iterative_consensus(sequences, program='muscle-medium', threshold=args.consensus_threshold/100)
			# If the sequence lengths are different, the beginning can be unclear
			cons = cons.lstrip('N')
		logger.info('Gene %s has %s assignments, %s usable (%s unique sequences). '
			'Consensus has %s N bases.', name, len(group), len(sequences),
			len(set(sequences)), cons.count('N'))
		if cons.count('N') > 0:
			n_consensus_with_n += 1
			if args.no_ambiguous:
				continue
		n_written += 1
		print('>{} {}_consensus\n{}'.format(name, args.part, cons))

	in_or_ex = 'excluding' if args.no_ambiguous else 'including'
	logger.info('Wrote a consensus for %s of %s genes (%s %s with ambiguous bases)', n_written, n_genes, in_or_ex, n_consensus_with_n)



"""
Species-specific code, such as lists of motifs and regular expressions.

Some refactoring is needed to make this module actually usable for many
species. Right now, it works for - at least - human, rhesus monkey and mouse.
"""
import re
from sqt.dna import amino_acid_regex
from .utils import nt_to_aa


# Regular expressions for CDR3 detection
_CDR3_REGEX = {
	# Heavy chain.
	#
	# This is a slightly improved version of the regular expression by
	# D’Angelo et al.: The antibody mining toolbox.
	# http://dx.doi.org/10.4161/mabs.27105
	# The amino-acid version of the expression is:
	# [FY][FWVHY]C[ETNGASDRIKVM]X{5,32}W[GAV]
	'VH': re.compile("""
		(TT[CT] | TA[CT])                            # F or Y
		(TT[CT] | TA[CT] | CA[CT] | GT[ACGT] | TGG)  # any of F, Y, H, V, W
		(TG[CT])                                     # C
		(?P<cdr3>                                    # actual CDR3 starts here
			(([GA][AGCT]) | TC | CG) [ACGT]          # any of ETNGASDRIKVM
			([ACGT]{3}){5,31}                        # between five and 31 codons
		)                                            # end of CDR3
		TGG                                          # W
		G[CGT][ACGT]                                 # G, A or V
		""", re.VERBOSE),


	# Light chain, kappa.
	# The amino-acid version is: [YFVS][YNHFCV][CWFGDLS]X{5,15}[FLV][GRV]
	'VK': re.compile("""
		(TA[CT] | TT[CT] | GT[ACGT] | TC[ACGT]|AG[CT] )           # Y, F, V, S
		(TA[CT] | AA[CT] | CA[CT] | TT[CT] | TG[CT] | GT[ACGT] )  # Y, N, H, F, C, V
		(TG[CT] | TGG | TT[CT] | GG[ACGT] | GA[CT] |              # C, W, F, G, D ...
			TC[ACGT] | AG[CT] |                                   # ... or S
			CT[ACGT] | TT[AG] )                                   # ... or L,
		(?P<cdr3>
			([ACGT]{3}){5,15}   # between five and fifteen codons
		)
		(
			TT[CT]   | GT[ACGT] |                                # F, V ...
			CT[ACGT] | TT[AG]                                    # ... or L
		)
		(
			GG[ACGT] | GT[ACGT] |                                # G, V ...
			CG[ACGT] | AG[AG]T                                   # ... or R
		)
		""", re.VERBOSE),


	# Light chain, lambda.
	# The amino-acid version is: [CDY][CFHSY][CFGW]X{5,15}[FS]G
	'VL': re.compile("""
		(TG[CT] | GA[CT] | TA[CT] )                             # C, D, Y
		(TG[CT] | TT[CT] | CA[CT] | TC[ACGT]|AG[CT] | TA[CT] )  # C, F, H, S, Y
		(TG[CT] | TT[CT] | GG[ACGT] | TGG )                     # C, F, G, W
		(?P<cdr3>
			([ACGT]{3}){5,15}   # between five and fifteen codons
		)
		( TT[CT] | TC[ACGT]|AG[CT] )                   # F, S
		GG[ACGT]                                       # G
		""", re.VERBOSE)
}

_CDR3_VH_ALTERNATIVE_REGEX = re.compile("""
		C
		(?P<cdr3> . [RK] .{3,30})
		[WF]G.G
""", re.VERBOSE)


def find_cdr3(sequence, chain):
	"""
	Find the CDR3 in the given sequence, assuming it comes from the given chain ('VH', 'VK', 'VL').
	If the chain is not one of 'VH', 'VK', 'VL', return None.

	Return a tuple (start, stop) if found, None otherwise.
	"""
	if chain not in _CDR3_REGEX:
		return None
	match = _CDR3_REGEX[chain].search(sequence)
	if match:
		return match.span('cdr3')
	if chain == 'VH':
		for offset in 0, 1, 2:
			aa = nt_to_aa(sequence[offset:])
			match = _CDR3_VH_ALTERNATIVE_REGEX.search(aa)
			if match:
				start, stop = match.span('cdr3')
				start = start * 3 + offset
				stop = stop * 3 + offset
				return start, stop
	return None


# When searching for the CDR3, start this many bases to the left of the end of
# the V match.
CDR3_SEARCH_START = 30


def _build_V_gene_regex():
	r = '('
	r += '|'.join(amino_acid_regex(aa) for aa in 'DV EA EM EV LV QV QL QE VE'.split())
	r += ')' + amino_acid_regex('Q')
	r += '([ACGT]{3})*'

	#r += '(' + amino_acid_regex('F') + '|' + amino_acid_regex('Y') + ')'
	#r += '([ACGT]{3})*'  # any codon
	# beginning of the CDR3 expression
	#r += '(TT[TC]|TA[CT])(TT[CT]|TA[TC]|CA[TC]|GT[AGCT]|TGG)(TG[TC])(([GA][AGCT])|TC|CG)[AGCT]'
	#r += '[ACGT]{4,6}$'
	return re.compile(r)


# TODO this is unused
_V_GENE_REGEX = _build_V_gene_regex()


# TODO this is work in progress
def looks_like_V_gene(s, chain):
	"""
	Check whether the given sequence matches our expectations of how a V gene
	should look like.
	"""
	s = s.upper()
	for start in START_MOTIFS[chain]:
		if s.startswith(start):
			return True #break
	else:
		return False

	# TODO unusued
	for end in 'TATTACTGT TTTTACTGT TATTACTGC TATTACTGC TATTGTGCA TATTACTGC TATTATTGT'.split():
		if s[-len(end)-13:].find(end) != -1:
			return True
	return False

	#return bool(_V_GENE_REGEX.match(s))


# A newly discovered V gene must start with one of these motifs
START_MOTIFS = {
	'VH': [
		'CAAAT',
		'CAGAT',
		'CAGCT',
		'CAGGA',
		'CAGGT',
		'CGGCT',
		'CGGGT',
		'CTGGT',
		'GAAAT',
		'GAAGT',
		'GACGT',
		'GAGAT',
		'GAGGC',
		'GAGGT',
		'GTGGA',
	],
	'VK': [
		'AACATC',
		'AATATT',
		'AATTTC',
		'CAAGTT',
		'CAGACT',
		'GAAACA',
		'GAAACG',
		'GAAATA',
		'GAAATG',
		'GAAATT',
		'GACATC',
		'GACATG',
		'GACATT',
		'GAGATT',
		'GATACT',
		'GATATC',
		'GATATT',
		'GATGCT',
		'GATGTT',
		'GATTTC',
		'GCCATC',
		'GCTGTT',
		'GTTATT',
	],
	'VL': [
		'AAGCCT',
		'AATTTT',
		'ACATAT',
		'AGGTCT',
		'CAAGCT',
		'CAATCT',
		'CACCCT',
		'CAGACT',
		'CAGCCT',
		'CAGGCT',
		'CAGTCT',
		'CGGTCT',
		'CTGACT',
		'CTGTCT',
		'GAGGCT',
		'GAGGTT',
		'GCATCT',
		'TCCACA',
		'TCCTAT',
	]
}

for _motif_list in START_MOTIFS.values():
	assert len(set(_motif_list)) == len(_motif_list)



"""
Filter table with parsed IgBLAST results

Discard the following rows in the table:
- no J assigned
- stop codon found
- V gene coverage less than 90
- J gene coverage less than 60
- V gene E-value greater than 1E-3

The filtered table is printed to standard output.
"""
import logging
import pandas as pd

from .table import read_table

logger = logging.getLogger(__name__)

def add_arguments(parser):
	parser.add_argument('table', help='Table with filtered IgBLAST results.')


def filtered_table(table,
		v_gene_coverage=90,  # at least
		j_gene_coverage=60,  # at least
		v_gene_evalue=1E-3,  # at most
	):
	"""
	Discard the following rows in the table (read in by read_table):
	- no J assigned
	- stop codon found
	- V gene coverage less than v_gene_coverage
	- J gene coverage less than j_gene_coverage
	- V gene E-value greater than v_gene_evalue

	Return the filtered table.
	"""
	# Both V and J must be assigned
	# (Note V_gene and J_gene columns use empty strings instead of NA)
	filtered = table[(table['V_gene'] != '') & (table['J_gene'] != '')][:]
	logger.info('%s rows have both V and J assignment', len(filtered))
	filtered['V_gene'] = pd.Categorical(filtered['V_gene'])

	# Filter out sequences that have a stop codon
	filtered = filtered[filtered.stop == 'no']
	logger.info('%s of those do not have a stop codon', len(filtered))

	# Filter out sequences with a too low V gene hit E-value
	filtered = filtered[filtered.V_evalue <= v_gene_evalue]
	logger.info('%s of those have an E-value of at most %s', len(filtered), v_gene_evalue)

	# Filter out sequences with too low V gene coverage
	filtered = filtered[filtered.V_covered >= v_gene_coverage]
	logger.info('%s of those cover the V gene by at least %s%%', len(filtered), v_gene_coverage)

	# Filter out sequences with too low J gene coverage
	filtered = filtered[filtered.J_covered >= j_gene_coverage]
	logger.info('%s of those cover the J gene by at least %s%%', len(filtered), j_gene_coverage)

	return filtered


def main(args):
	d = read_table(args.table, log=True)
	d = filtered_table(d)
	print(d.to_csv(sep='\t', index=False), end='')
	logger.info('%d rows written', len(d))



#!/usr/bin/env python3
"""
IgDiscover computes V/D/J gene usage profiles and discovers novel V genes

- Run IgBLAST in parallel (wrapper inspired by igblastwrp).
- Parse IgBLAST output into a tab-separated table
- Group sequences by barcode
- Plot V gene usage
- Discover new V genes given more than one dataset
"""
__author__ = "Marcel Martin"

import logging
import importlib
from sqt import HelpfulArgumentParser
import matplotlib as mpl
import warnings
from . import __version__

mpl.use('Agg')
warnings.filterwarnings('ignore', 'axes.color_cycle is deprecated and replaced with axes.prop_cycle')
warnings.filterwarnings('ignore', 'The `IPython.html` package')

# List of all subcommands. A module of the given name must exist and define
# add_arguments() and main() functions. Documentation is taken from the first
# line of the module’s docstring.
COMMANDS = [
	'init',
	'run',
	'commonv',
	'igblast',
	'parse',
	'filter',
	'count',
	'group',
	#'multidiscover',
	'compose',
	'discover',
	'clusterplot',
	'errorplot',
	'upstream',
	'dendrogram',
	'rename',
	'union',
]

logger = logging.getLogger(__name__)

def main():
	logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
	parser = HelpfulArgumentParser(description=__doc__, prog='igdiscover')
	parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

	subparsers = parser.add_subparsers()
	for command_name in COMMANDS:
		module = importlib.import_module('.' + command_name, 'igdiscover')
		subparser = subparsers.add_parser(command_name,
			help=module.__doc__.split('\n')[1], description=module.__doc__)
		subparser.set_defaults(func=module.main)
		module.add_arguments(subparser)

	args = parser.parse_args()
	if not hasattr(args, 'func'):
		parser.error('Please provide the name of a subcommand to run')
	else:
		args.func(args)


if __name__ == '__main__':
	main()



"""
Parse IgBLAST output and write out a tab-separated table.

IgBLAST must have been run with -outfmt "7 sseqid qstart qseq sstart sseq pident slen"

A few extra things are done in addition to parsing:
- The CDR3 is detected by using a regular expression
- The leader is detected within the sequence before the found V gene (by
  searching for the start codon).
- If the V sequence hit starts at base 2 in the reference, it is extended
  one to the left.
"""
import re
import sys
import csv
import errno
import logging
from collections import namedtuple

import pandas as pd
from sqt import SequenceReader, xopen
from sqt.dna import reverse_complement
from .utils import nt_to_aa
from .species import find_cdr3, CDR3_SEARCH_START

logger = logging.getLogger(__name__)


def add_arguments(parser):
	parser.add_argument('--rename', default=None, metavar='PREFIX',
		help='Rename reads to PREFIXseqN (where N is a number starting at 1)')
	parser.add_argument('--vdatabase', '--vdb', metavar='FASTA',
		help="Path to FASTA file with V genes. Used to fix the 5' ends of V "
		"gene alignments. If not given, 'N' bases will be inserted instead.")
	parser.add_argument('--hdf5', metavar='FILE',
		help='Write table in HDF5 format to FILE')
	parser.add_argument('igblast', help='IgBLAST output')
	parser.add_argument('fasta', help='File with original reads')


def none_if_na(s):
	"""Return None if s == 'N/A'. Return s otherwise."""
	return None if s == 'N/A' else s


def split_by_section(it, section_starts):
	"""
	Parse a stream of lines into chunks of sections. When one of the lines
	starts with a string given in section_starts, a new section is started, and
	a tuple (head, lines) is returned where head is the matching line and lines
	contains a list of the lines following the section header, up to (but
	excluding) the next section header.

	Works a bit like str.split(), but on lines.
	"""
	lines = None
	header = None
	for line in it:
		line = line.strip()
		for start in section_starts:
			if line.startswith(start):
				if header is not None:
					yield (header, lines)
				header = line
				lines = []
				break
		else:
			if header is None:
				raise ParseError("Expected a line starting with one of {}".format(', '.join(section_starts)))
			lines.append(line)
	if header is not None:
		yield (header, lines)


AlignmentSummary = namedtuple('AlignmentSummary', 'start stop length matches mismatches gaps percent_identity')
JunctionVDJ = namedtuple('JunctionVDJ', 'v_end vd_junction d_region dj_junction j_start')
JunctionVJ = namedtuple('JunctionVJ', 'v_end vj_junction j_start')


_Hit = namedtuple('_Hit', [
	'subject_id',  # name of database record, such as "VH4.11"
	'query_start',
	'query_sequence',  # aligned part of the query
	'subject_start',
	'subject_sequence',  # aligned part of the reference
	'subject_length',  # total length of reference, depends only on subject_id
	'errors',
	'percent_identity',
	'evalue',
])
class Hit(_Hit):
	# This avoids having a __dict__ attribute, which is necessary for namedtuple
	# subclasses that need _asdict() to work (http://bugs.python.org/issue24931)
	__slots__ = ()

	def covered(self):
		"""
		Return fraction of bases in the original subject sequence that are
		covered by this hit.
		"""
		return len(self.subject_sequence) / self.subject_length

	@property
	def query_end(self):
		return self.query_start + len(self.query_sequence)


def parse_header(header):
	"""
	Extract size= and barcode= fields from the FASTA/FASTQ header line

	>>> parse_header("name;size=12;barcode=ACG;")
	('name', 12, 'ACG')
	>>> parse_header("another name;size=200;foo=bar;")
	('another name', 200, None)
	"""
	fields = header.split(';')
	query_name = fields[0]
	size = barcode = None
	for field in fields[1:]:
		if field == '':
			continue
		if '=' in field:
			key, value = field.split('=', maxsplit=1)
			if key == 'size':
				size = int(value)
			elif key == 'barcode':
				barcode = value
	return query_name, size, barcode


IgBlastRecord = namedtuple('IgBlastRecord',
	'full_sequence query_name alignments '
	'hits v_gene d_gene j_gene chain has_stop in_frame is_productive '
	'strand junction')


class ExtendedIgBlastRecord(IgBlastRecord):
	"""
	This extended record does a few extra things:
	- The CDR3 is detected by using a regular expression
	- The leader is detected within the sequence before the found V gene (by
	searching for the start codon).
	- The RACE-specific run of G in the beginning of the sequence is detected.
	- If the V sequence hit starts at base 2 in the reference, it is extended
	  one to the left.
	"""
	# TODO move computation of cdr3_sequence, vdj_sequence into constructor
	# TODO maybe make all coordinates relative to full sequence

	# Order of columns (use with asdict())
	columns = [
		'count',
		'V_gene',
		'D_gene',
		'J_gene',
		'chain',
		'stop',
		'productive',
		'V_covered',
		'D_covered',
		'J_covered',
		'V_evalue',
		'D_evalue',
		'J_evalue',
		'FR1_SHM',
		'CDR1_SHM',
		'FR2_SHM',
		'CDR2_SHM',
		'FR3_SHM',
		'V_SHM',
		'J_SHM',
		'V_errors',
		'J_errors',
		'UTR',
		'leader',
		'CDR1_nt',
		'CDR1_aa',
		'CDR2_nt',
		'CDR2_aa',
		'CDR3_nt',
		'CDR3_aa',
		'V_nt',
		'V_aa',
		'V_end',
		'VD_junction',
		'D_region',
		'DJ_junction',
		'J_nt',
		'VDJ_nt',
		'VDJ_aa',
		'name',
		'barcode',
		'race_G',
		'genomic_sequence',
	]
	def __new__(cls, record, v_database):
		d = record._asdict()
		query_name, size, barcode = parse_header(record.query_name)
		d['query_name'] = query_name
		obj = super().__new__(cls, **d)
		obj.size = size
		obj.barcode = barcode
		return obj

	def __init__(self, record, v_database):
		self.genomic_sequence = self.full_sequence
		self.race_g = None  # TODO since the parse script does not extract the race_G anymore, we don’t have this info available
		if 'V' in self.hits:
			self.hits['V'] = self._fixed_v_hit(v_database)
		self.utr, self.leader = self._utr_leader()
		self.alignments['CDR3'] = self._fixed_cdr3_alignment()

	@property
	def vdj_sequence(self):
		if 'V' not in self.hits or 'J' not in self.hits:
			return None
		hit_v = self.hits['V']
		hit_j = self.hits['J']
		vdj_start = hit_v.query_start
		vdj_stop = hit_j.query_start + len(hit_j.query_sequence)
		return self.full_sequence[vdj_start:vdj_stop]

	def _utr_leader(self):
		"""
		Split the sequence before the V gene match into UTR and leader by
		searching for the start codon.
		"""
		if 'V' not in self.hits:
			return None, None
		before_v = self.full_sequence[:self.hits['V'].query_start]

		# Search for the start codon
		for offset in (0, 1, 2):
			for i in range(66, 42, -3):
				if before_v[-i + offset : -i + 3 + offset] == 'ATG':
					return before_v[:-i + offset], before_v[-i + offset:]
		return None, None

	def _fixed_cdr3_alignment(self):
		"""
		Return a repaired AlignmentSummary object for the CDR3 region which
		does not use IgBLAST’s coordinates. IgBLAST does not determine the end
		of the CDR3 correctly, at least when a custom database is used,

		Return (start, end) of CDR3 relative to query. The CDR3 is detected
		using a regular expression. Return None if no CDR3 detected.
		"""
		if 'V' not in self.hits or 'J' not in self.hits:
			return None
		# Search in a window around the V(D)J junction for the CDR3
		if 'CDR3' in self.alignments:
			window_start = self.alignments['CDR3'].start - CDR3_SEARCH_START
		else:
			window_start = max(0, self.hits['V'].query_end - CDR3_SEARCH_START)
		window_end = self.hits['J'].query_end
		window = self.full_sequence[window_start:window_end]
		match = find_cdr3(window, self.chain)
		if not match:
			return None
		start = match[0] + window_start
		end = match[1] + window_start
		assert start < end
		return AlignmentSummary(start=start, stop=end, length=None, matches=None,
			mismatches=None, gaps=None, percent_identity=None)

	def _fixed_v_hit(self, v_database):
		"""
		Extend the V hit by one base to the left if it starts at the second
		base in the V reference.
		"""
		hit = self.hits['V']
		if hit.subject_start != 1 or hit.query_start == 0:
			return hit
		d = hit._asdict()
		assert self.full_sequence[d['query_start']:d['query_start'] + len(d['query_sequence'])] == d['query_sequence']
		d['query_start'] -= 1
		d['subject_start'] -= 1
		d['query_sequence'] = self.full_sequence[d['query_start']:d['query_start'] + len(d['query_sequence']) + 1]
		if v_database:
			reference = v_database[hit.subject_id]
			preceding_base = reference[d['subject_start']]
		else:
			preceding_base = 'N'
		d['subject_sequence'] = preceding_base + d['subject_sequence']  # TODO maybe not needed
		# TODO one could adjust the no. of errors, percent_identity (and E-value)
		return Hit(**d)

	def region_sequence(self, region):
		"""
		Return sequence of a named region. Allowed names are:
		CDR1, CDR2, CDR3, FR1, FR2, FR3. For all regions except CDR3, sequences
		are extracted from the full read using begin and end coordinates from
		IgBLAST’s "alignment summary" table.
		"""
		alignment = self.alignments.get(region, None)
		if alignment is None:
			return None
		if alignment.start is None or alignment.stop is None:
			return None
		return self.full_sequence[alignment.start:alignment.stop]

	def asdict(self):
		"""
		Return a flattened representation of this record as a dictionary.
		The dictionary can then be used with e.g. a csv.DictWriter or
		pandas.DataFrame.from_items.
		"""
		cdr1nt = self.region_sequence('CDR1')
		cdr1aa = nt_to_aa(cdr1nt) if cdr1nt else None
		cdr2nt = self.region_sequence('CDR2')
		cdr2aa = nt_to_aa(cdr2nt) if cdr2nt else None
		cdr3nt = self.region_sequence('CDR3')
		cdr3aa = nt_to_aa(cdr3nt) if cdr3nt else None
		vdj_nt = self.vdj_sequence
		vdj_aa = nt_to_aa(vdj_nt) if vdj_nt else None

		def shm(region):
			if region in self.alignments:
				rar = self.alignments[region]
				if rar is None or rar.percent_identity is None:
					return None
				return 100. - rar.percent_identity
			else:
				return None

		if 'V' in self.hits:
			v_nt = self.hits['V'].query_sequence
			v_aa = nt_to_aa(v_nt)
			v_shm = 100. - self.hits['V'].percent_identity
			v_errors = self.hits['V'].errors
			v_covered = 100. * self.hits['V'].covered()
			v_evalue = self.hits['V'].evalue
		else:
			v_nt = None
			v_aa = None
			v_shm = None
			v_errors = None
			v_covered = None
			v_evalue = None
		if 'D' in self.hits:
			d_covered = 100. * self.hits['D'].covered()
			d_evalue = self.hits['D'].evalue
		else:
			d_covered = None
			d_evalue = None
		if 'J' in self.hits:
			j_nt = self.hits['J'].query_sequence
			j_shm = 100. - self.hits['J'].percent_identity
			j_errors = self.hits['J'].errors
			j_covered = 100. * self.hits['J'].covered()
			j_evalue = self.hits['J'].evalue
		else:
			j_nt = None
			j_shm = None
			j_errors = None
			j_covered = None
			j_evalue = None
		v_end = getattr(self.junction, 'v_end', None)
		vd_junction = getattr(self.junction, 'vd_junction', None)
		d_region = getattr(self.junction, 'd_region', None)
		dj_junction = getattr(self.junction, 'dj_junction', None)
		return dict(
			count=self.size,
			V_gene=self.v_gene,
			D_gene=self.d_gene,
			J_gene=self.j_gene,
			chain=self.chain,
			stop=self.has_stop,
			productive=self.is_productive,
			V_covered=v_covered,
			D_covered=d_covered,
			J_covered=j_covered,
			V_evalue=v_evalue,
			D_evalue=d_evalue,
			J_evalue=j_evalue,
			FR1_SHM=shm('FR1'),
			CDR1_SHM=shm('CDR1'),
			FR2_SHM=shm('FR2'),
			CDR2_SHM=shm('CDR2'),
			FR3_SHM=shm('FR3'),
			V_SHM=v_shm,
			J_SHM=j_shm,
			V_errors=v_errors,
			J_errors=j_errors,
			UTR=self.utr,
			leader=self.leader,
			CDR1_nt=cdr1nt,
			CDR1_aa=cdr1aa,
			CDR2_nt=cdr2nt,
			CDR2_aa=cdr2aa,
			CDR3_nt=cdr3nt,
			CDR3_aa=cdr3aa,
			V_nt=v_nt,
			V_aa=v_aa,
			V_end=v_end,
			VD_junction=vd_junction,
			D_region=d_region,
			DJ_junction=dj_junction,
			J_nt=j_nt,
			VDJ_nt=vdj_nt,
			VDJ_aa=vdj_aa,
			name=self.query_name,
			barcode=self.barcode,
			race_G=self.race_g,
			genomic_sequence=self.genomic_sequence,
		)


class ParseError(Exception):
	pass


class IgBlastParser:
	"""
	Parser for IgBLAST results. Works only when IgBLAST was run with
	the option -outfmt "7 sseqid qstart qseq sstart sseq pident slen".
	"""
	BOOL = { 'Yes': True, 'No': False, 'N/A': None }
	FRAME = { 'In-frame': True, 'Out-of-frame': False, 'N/A': None }
	SECTIONS = frozenset([
		'# Query:',
		'# V-(D)-J rearrangement summary',
		'# V-(D)-J junction details',
		'# Alignment summary',
		'# Hit table',
	])

	def __init__(self, sequence_file, igblast_file):
		self._sequence_reader = SequenceReader(sequence_file)
		self._igblast_file = igblast_file

	def __iter__(self):
		"""
		Yield IgBlastRecord objects
		"""
		zipped = zip(self._sequence_reader, split_by_section(self._igblast_file, ['# IGBLASTN']))
		for fasta_record, (record_header, record_lines) in zipped:
			assert record_header == '# IGBLASTN 2.2.29+'
			yield self._parse_record(record_lines, fasta_record)

	def _parse_record(self, record_lines, fasta_record):
		"""
		Parse a single IgBLAST record
		"""
		hits = dict()
		# All of the sections are optional, so we need to set default values here.
		query_name = None
		junction = None
		v_gene, d_gene, j_gene, chain, has_stop, in_frame, is_productive, strand = [None] * 8
		alignments = dict()
		for section, lines in split_by_section(record_lines, self.SECTIONS):
			if section.startswith('# Query: '):
				query_name = section.split(': ')[1]
			elif section.startswith('# V-(D)-J rearrangement summary'):
				fields = lines[0].split('\t')
				if len(fields) == 7:
					# No D assignment
					v_gene, j_gene, chain, has_stop, in_frame, is_productive, strand = fields
					d_gene = None
				else:
					v_gene, d_gene, j_gene, chain, has_stop, in_frame, is_productive, strand = fields
				v_gene = none_if_na(v_gene)
				d_gene = none_if_na(d_gene)
				j_gene = none_if_na(j_gene)
				chain = none_if_na(chain)
				has_stop = self.BOOL[has_stop]
				in_frame = self.FRAME[in_frame]
				is_productive = self.BOOL[is_productive]
				strand = strand if strand in '+-' else None
			elif section.startswith('# V-(D)-J junction details'):
				fields = lines[0].split('\t')
				if len(fields) == 5:
					junction = JunctionVDJ(
						v_end=fields[0],
						vd_junction=fields[1],
						d_region=fields[2],
						dj_junction=fields[3],
						j_start=fields[4]
					)
				else:
					junction = JunctionVJ(
						v_end=fields[0],
						vj_junction=fields[1],
						j_start=fields[2])
			elif section.startswith('# Alignment summary'):
				for line in lines:
					fields = line.split('\t')
					if len(fields) == 8 and fields[0] != 'Total':
						summary = self._parse_alignment_summary(fields[1:])
						region_name, _, imgt = fields[0].partition('-')
						assert imgt in ('IMGT', 'IMGT (germline)')
						alignments[region_name] = summary
			elif section.startswith('# Hit table'):
				for line in lines:
					if not line or line.startswith('#'):
						continue
					hit, gene = self._parse_hit(line)
					assert gene in ('V', 'D', 'J')
					assert gene not in hits, "Two hits for same gene found"
					hits[gene] = hit

		assert fasta_record.name == query_name
		full_sequence = fasta_record.sequence.upper()
		if strand == '-':
			full_sequence = reverse_complement(full_sequence)

		if __debug__:
			for gene in ('V', 'D', 'J'):
				if gene not in hits:
					continue
				hit = hits[gene]

				qsequence = hit.query_sequence
				#print(gene, hit.query_start, '-', hit.query_start + len(qsequence), end=' ')

				# IgBLAST removes the trailing semicolon (why, oh why??)
				qname = query_name[:-1] if query_name.endswith(';') else query_name
				assert chain in (None, 'VL', 'VH', 'VK', 'NON'), chain
				assert qsequence == full_sequence[hit.query_start:hit.query_start+len(qsequence)]

		return IgBlastRecord(
			query_name=query_name,
			alignments=alignments,
			v_gene=v_gene,
			d_gene=d_gene,
			j_gene=j_gene,
			chain=chain,
			has_stop=has_stop,
			in_frame=in_frame,
			is_productive=is_productive,
			strand=strand,
			hits=hits,
			full_sequence=full_sequence,
			junction=junction)

	def _parse_alignment_summary(self, fields):
		start, stop, length, matches, mismatches, gaps = (int(v) for v in fields[:6])
		percent_identity = float(fields[6])
		return AlignmentSummary(
			start=start - 1,
			stop=stop,
			length=length,
			matches=matches,
			mismatches=mismatches,
			gaps=gaps,
			percent_identity=percent_identity
		)


	def _parse_hit(self, line):
		"""
		Parse a line of the "Hit table" section and return a tuple (hit, gene)
		where hit is a Hit object.
		"""
		gene, subject_id, query_start, query_sequence, subject_start, subject_sequence, percent_identity, subject_length, evalue = line.split('\t')
		# subject_sequence and query_sequence describe the alignment:
		# They contain '-' characters for insertions and deletions.
		assert len(subject_sequence) == len(query_sequence)
		alignment_length = len(subject_sequence)
		errors = sum(a != b for a,b in zip(subject_sequence, query_sequence))
		query_sequence = query_sequence.replace('-', '')
		subject_sequence = subject_sequence.replace('-', '')
		query_start = int(query_start) - 1
		subject_start = int(subject_start) - 1
		subject_length = int(subject_length)  # Length of original subject sequence
		#subject_end = subject_start + len(subject_sequence)
		# Percent identity is calculated by IgBLAST as
		# 100 - errors / alignment_length and then rounded to two decimal digits
		percent_identity = float(percent_identity)
		evalue = float(evalue)
		hit = Hit(subject_id, query_start, query_sequence, subject_start,
			subject_sequence, subject_length, errors, percent_identity, evalue)
		return hit, gene


class TableWriter:
	def __init__(self, file):
		"""
		If rename is not None, rename reads to {rename}seq{number}, where
		{number} is a sequential number starting from 1.
		"""
		self._file = file
		self._writer = csv.DictWriter(file, fieldnames=ExtendedIgBlastRecord.columns, delimiter='\t')
		self._writer.writeheader()

	@staticmethod
	def yesno(v):
		"""
		Return "yes", "no" or None for boolean value v, which may also be None.
		"""
		if v is None:
			return None
		return ["no", "yes"][v]

	def write(self, d):
		"""
		Write the IgBLAST record (must be given as dictionary) to the output
		file.
		"""
		d = d.copy()
		d['stop'] = self.yesno(d['stop'])
		d['productive'] = self.yesno(d['productive'])
		for name in ('V_covered', 'D_covered', 'J_covered',
				'FR1_SHM', 'CDR1_SHM', 'FR2_SHM', 'CDR2_SHM', 'FR3_SHM',
				'V_SHM', 'J_SHM'):
			if d[name] is not None:
				d[name] = '{:.1f}'.format(d[name])
		for name in ('V_evalue', 'D_evalue', 'J_evalue'):
			if d[name] is not None:
				d[name] = '{:G}'.format(d[name])
		self._writer.writerow(d)


def main(args):
	"""
	Parse IgBLAST output
	"""
	n = 0
	if args.hdf5:
		rows = []

	if args.vdatabase:
		with SequenceReader(args.vdatabase) as fr:
			v_database = { record.name: record.sequence.upper() for record in fr }
	else:
		v_database = None

	cdr3_recognized = 0
	writer = TableWriter(sys.stdout)
	with xopen(args.fasta) as fasta, xopen(args.igblast) as igblast:
		parser = IgBlastParser(fasta, igblast)
		for record in parser:
			n += 1
			extended_record = ExtendedIgBlastRecord(record, v_database=v_database)
			d = extended_record.asdict()
			if d['CDR3_aa']:
				cdr3_recognized += 1
			if args.rename is not None:
				d['name'] = "{}seq{}".format(args.rename, n)
			if args.hdf5:
				rows.append(d)
			try:
				writer.write(d)
			except IOError as e:
				if e.errno == errno.EPIPE:
					sys.exit(1)
				raise
	logger.info('%d records parsed and written', n)
	logger.info('CDR3s detected in %.1f%% of all sequences', cdr3_recognized / n * 100)
	if args.hdf5:
		from igdiscover.table import STRING_COLUMNS, INTEGER_COLUMNS
		df = pd.DataFrame(rows, columns=ExtendedIgBlastRecord.columns)
		# TODO code below is copied from igdiscover.table
		# Convert all string columns to str to avoid a PerformanceWarning
		for col in STRING_COLUMNS:
			df[col].fillna('', inplace=True)
			df[col] = df[col].astype('str')
			# Empty strings have been set to NaN by read_csv. Replacing
			# by the empty string avoids problems with groupby, which
			# ignores NaN values.
			# Columns that have any NaN values in them cannot be converted to
			# int due to a numpy limitation.
			for col in INTEGER_COLUMNS:
				if all(df[col].notnull()):
					df[col] = df[col].astype(int)
		df.to_hdf(args.hdf5, 'table', mode='w', complevel=3, complib='zlib')
		logger.info('HDF5 file written')



"""
"""
from igdiscover.parse import parse_header
from nose.tools import raises


def test():
	assert parse_header('') == ('', None, None)
	assert parse_header('abc') == ('abc', None, None)
	assert parse_header('abc;size=17;') == ('abc', 17, None)
	assert parse_header('abc;barcode=ACG') == ('abc', None, 'ACG')
	assert parse_header('abc;size=17;barcode=ACG') == ('abc', 17, 'ACG')
	assert parse_header('abc;size=17;barcode=ACG;soze=19;baarcode=GGG') == ('abc', 17, 'ACG')



"""
Group sequences that share a barcode (molecular identifier, MID)

Since the same barcode can sometimes be used by different sequences, a 'pseudo
CDR3' sequence (defined to be 80 to 60 bases upstream of the 3' end) is used
to further distinguish sequences.

The barcode is assumed to be in the 5' end of the sequence.

Use --trim-g to remove initial runs of G (at 5' end, artifact from RACE
protocol).

For all the found groups, one sequence is output to standard output (in FASTA
format). Which sequence that is depends on the group size:
- If the group consists of a single sequence, that sequence is output
- If the group consists of two sequences, one of them is picked randomly
- If the group has at least three sequences, a consensus is computed. The
  consensus is output if it contains no ambiguous bases. Otherwise, also here a
  random sequence is chosen.
"""
"""
Things to keep in mind
- There are some indels in homopolymers (probably PCR problem)
- Different lengths of the initial G run cannot be used to distinguish sequences
  since they can come from polymerase problems in homopolymers.
- There are also regular sequencing errors in the initial run of G nucleotides.
- Some paired reads aren’t correctly merged into single reads. They end up being
  too long.
- When grouping by barcode and pseudo CDR3, sequence lengths vary within groups.
  However, this affects only ~1% of sequences, so it is not necessary to compute
  a multiple alignment. Just taking the consensus will drown the incorrect
  sequences, at least if the group size is large.
- It does not hurt to reduce the minimimum number of sequences per group for
  taking a consensus to 2, but it also does not help much (results in 0.5% more
  sequences). (The consensus is only successful if both sequences are identical.)
  However, since this is also a simple way to deal with exact duplicates, we do
  it anyway and can then skip the separate duplicate removal step (VSEARCH).

TODO
- Use pandas.DataFrame
"""

import csv
import sys
import logging
from collections import Counter, defaultdict
from contextlib import ExitStack
from itertools import islice

import matplotlib
matplotlib.use('pdf')
import matplotlib.pyplot as plt
from sqt.align import consensus
from sqt import SequenceReader

from .species import find_cdr3

# minimum number of sequences needed for attempting to compute a consensus
MIN_CONSENSUS_SEQUENCES = 3


logger = logging.getLogger(__name__)

def add_arguments(parser):
	parser.add_argument('--groups-output', metavar='FILE', default=None,
		help='Write tab-separated table with groups to FILE')
	parser.add_argument('--plot-sizes', metavar='FILE', default=None,
		help='Plot group sizes to FILE (.png or .pdf)')
	parser.add_argument('--real-cdr3', action='store_true', default=False,
		help='Use the real CDR3 (detected with regex) instead of pseudo CDR3')
	parser.add_argument('--limit', default=None, type=int, metavar='N',
		help='Limit processing to the first N reads')
	parser.add_argument('--trim-g', action='store_true', default=False,
		help="Trim 'G' nucleotides at 5' end")
	parser.add_argument('--minimum-length', '-l', type=int, default=0,
		help='Minimum sequence length')
	parser.add_argument('--barcode-length', '-b', type=int, default=12,
		help="Length of barcode. Positive for 5' barcode, negative for 3' barcode. (default: %(default)s")
	parser.add_argument('fastx', metavar='FASTA/FASTQ',
		help='FASTA or FASTQ file (can be gzip-compressed) with sequences')


class Graph:
	"""Graph that can find connected components"""
	def __init__(self, nodes):
		self._nodes = { node: [] for node in nodes }

	def add_edge(self, node1, node2):
		self._nodes[node1].append(node2)
		self._nodes[node2].append(node1)

	def connected_components(self):
		"""Return a list of connected components."""
		visited = set()
		components = []
		for node, neighbors in self._nodes.items():
			if node in visited:
				continue
			# Start a new component
			to_visit = [node]
			component = []
			while to_visit:
				n = to_visit.pop()
				if n in visited:
					continue
				visited.add(n)
				component.append(n)
				for neighbor in self._nodes[n]:
					if neighbor not in visited:
						to_visit.append(neighbor)
			components.append(component)
		return components


def hamming_neighbors(s):
	"""Return sequences that are at hamming distance 1 and return also s itself"""
	for i in range(len(s)):
		for c in 'ACGT':
			if s[i] != c:
				yield s[:i] + c + s[i+1:]
	yield s


def cluster_sequences(sequences):
	"""
	Single-linkage clustering. Two sequences are linked if
	- their (pseudo-) CDR3 sequences have a hamming distance of at most 1
	- and their lengths differs by at most 2.
	"""
	graph = Graph(sequences)
	cdr3_seqs = defaultdict(list)
	for s in sequences:
		cdr3_seqs[s.cdr3].append(s)
	for sequence in sequences:
		for neighbor in hamming_neighbors(sequence.cdr3):
			neighbor_seqs = cdr3_seqs.get(neighbor, [])
			for neighbor_seq in neighbor_seqs:
				if abs(len(sequence) - len(neighbor_seq)) <= 2:
					graph.add_edge(sequence, neighbor_seq)
	components = graph.connected_components()
	assert sum(len(component) for component in components) == len(sequences)
	return components


GROUPS_HEADER = ['barcode', 'cdr3', 'name', 'sequence']

def write_group(csvfile, barcode, sequences):
	for sequence in sequences:
		row = [barcode, sequence.cdr3, sequence.name.split(maxsplit=1)[0], sequence.sequence]
		csvfile.writerow(row)
	csvfile.writerow([])


def main(args):
	if args.barcode_length == 0:
		sys.exit("Barcode length must be non-zero")

	# Map barcodes to lists of sequences
	barcodes = defaultdict(list)
	n = 0
	too_short = 0
	cdr3s = set()
	regex_fail = 0
	with SequenceReader(args.fastx) as f:
		for record in islice(f, 0, args.limit):
			if len(record) < args.minimum_length:
				too_short += 1
				continue
			if args.barcode_length > 0:
				barcode = record.sequence[:args.barcode_length]
				unbarcoded = record[args.barcode_length:]
			else:
				barcode = record.sequence[args.barcode_length:]
				unbarcoded = record[:args.barcode_length]

			if args.trim_g:
				# The RACE protocol leads to a run of non-template Gs in the beginning
				# of the sequence, after the barcode.
				unbarcoded.sequence = unbarcoded.sequence.lstrip('G')
				if unbarcoded.qualities:
					unbarcoded.qualities = unbarcoded.qualities[-len(unbarcoded.sequence):]

			if args.real_cdr3:
				match = find_cdr3(unbarcoded.sequence, chain='VH')
				if match:
					cdr3 = unbarcoded.sequence[match[0]:match[1]]
				else:
					regex_fail += 1
					continue
			else:
				cdr3 = unbarcoded.sequence[-80:-60]
			unbarcoded.cdr3 = cdr3  # TODO slight abuse of Sequence objects
			barcodes[barcode].append(unbarcoded)
			cdr3s.add(cdr3)
			n += 1

	logger.info('%s sequences in input', n + too_short + regex_fail)
	logger.info('%s sequences long enough', n + regex_fail)
	if args.real_cdr3:
		logger.info('Using the real CDR3')
		logger.info('%s times (%.2f%%), the CDR3 regex matched', n, n / (n + regex_fail) * 100)
	else:
		logger.info('Using the pseudo CDR3')
	logger.info('%s unique CDR3s', len(cdr3s))
	logger.info('%s unique barcodes', len(barcodes))
	barcode_singletons = sum(1 for seqs in barcodes.values() if len(seqs) == 1)
	logger.info('%s barcodes used by only a single sequence (singletons)', barcode_singletons)

	with ExitStack() as stack:
		if args.groups_output:
			group_out = csv.writer(stack.enter_context(
				open(args.groups_output, 'w')), delimiter='\t', lineterminator='\n')
			group_out.writerow(GROUPS_HEADER)
		else:
			group_out = None
		too_few = 0
		n_clusters = 0
		n_singletons = 0
		n_consensus = 0
		n_ambiguous = 0
		sizes = []
		for barcode, sequences in barcodes.items():
			if len(sequences) != len(set(s.name for s in sequences)):
				logger.error('Duplicate sequence records detected')
				sys.exit(1)
			clusters = cluster_sequences(sequences)
			n_clusters += len(clusters)
			for cluster in clusters:
				sizes.append(len(cluster))
				if group_out:
					write_group(group_out, barcode, cluster)
				if len(cluster) == 1:
					n_singletons += 1
				if len(cluster) < MIN_CONSENSUS_SEQUENCES:
					too_few += 1
					sequence = cluster[0].sequence
					name = cluster[0].name
					cdr3 = cluster[0].cdr3
				else:
					cons = consensus({s.name: s.sequence for s in cluster}, threshold=0.501)
					if 'N' in cons:
						# Pick the first sequence as the output sequence
						sequence = cluster[0].sequence
						name = cluster[0].name
						cdr3 = cluster[0].cdr3
						n_ambiguous += 1
					else:
						sequence = cons
						n_consensus += 1
						cdr3 = Counter(cl.cdr3 for cl in cluster).most_common(1)[0][0]
						name = 'consensus{}'.format(n_consensus)

				name = name.split(maxsplit=1)[0]
				if name.endswith(';'):
					name = name[:-1]
				print('>{};barcode={};cdr3={};size={};\n{}'.format(name, barcode,
					cdr3, len(cluster), sequence))

	logger.info('%d clusters (%d singletons)', n_clusters, n_singletons)
	logger.info('%d consensus sequences computed (from groups that had at least %d sequences)',
		n_consensus + n_ambiguous, MIN_CONSENSUS_SEQUENCES)
	logger.info('%d of those had no ambiguous bases', n_consensus)

	if args.plot_sizes:
		matplotlib.rcParams.update({'font.size': 14})
		fig = plt.figure()
		ax = fig.gca()
		v, _, _ = ax.hist(sizes, bins=100)
		ax.set_ylim(0, v[1:].max()*1.1)
		ax.set_xlabel('Group size')
		ax.set_ylabel('Read frequency')
		ax.set_title('Histogram of group sizes (>1)')
		ax.grid(axis='x')
		ax.tick_params(direction="outward", top=False, right=False)
		fig.set_tight_layout(True)
		fig.savefig(args.plot_sizes)
		logger.info('Plotted group sizes to %r', args.plot_sizes)

	if args.groups_output:
		logger.info('Groups written to %r', args.groups_output)



from io import StringIO
import pkg_resources
from igdiscover.utils import has_stop
from igdiscover.utils import Config


def test_has_stop():
	assert has_stop('TAA')
	assert has_stop('TAG')
	assert has_stop('TGA')
	assert has_stop('GGGTGA')

	assert not has_stop('GGG')
	assert not has_stop('TAT')
	assert not has_stop('GGGT')
	assert not has_stop('GGGTA')
	assert not has_stop('TATTG')


def test_config():
	empty_config = Config(file=StringIO('{}'))
	packaged_config = Config(file=pkg_resources.resource_stream('igdiscover', Config.DEFAULT_PATH))
	# force library name to be equal since it is dynamically determined
	empty_config.library_name = packaged_config.library_name = 'nolib'
	e = empty_config.__dict__
	p = packaged_config.__dict__
	assert e.keys() == p.keys()
	for k in e:
		assert e[k] == p[k], '{}: {} vs {}'.format(k, e[k], p[k])
	# assert empty_config == packaged_config



"""
Rename sequences in a target FASTA file using a template FASTA file

The sequences in the target file will get the name that they have in the
template file. Sequences are considered to be equivalent if one is a prefix of
the other.
"""
import logging
from sqt import FastaReader
from .utils import natural_sort_key

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--no-sort', dest='sort', action='store_false', default=True,
		help='Do not sort sequences by name')
	arg('--not-found', metavar='TEXT', default=' (not found)',
		help='Append this text to the record name when the sequence was not found '
		'in the template. Default: %(default)r')
	arg('template', help='FASTA file with correctly named sequences')
	arg('target', help='FASTA file with to-be renamed sequences')


class PrefixDict:
	"""
	A dict that maps strings to values, but where a prefix of a key is enough
	to retrieve the value.
	"""
	def __init__(self, items):
		self._items = []
		for k, v in items:
			self.add(k, v)

	def add(self, k, v):
		try:
			self[k]
			raise ValueError('Key {!r} already exists'.format(k))
		except KeyError:
			self._items.append((k, v))

	def __getitem__(self, key):
		found = None
		for seq, value in self._items:
			if seq.startswith(key) or key.startswith(seq):
				if found is not None:
					# TODO don't use keyerror here
					raise KeyError('Key {!r} is ambiguous'.format(key))
				found = value
		if found is None:
			raise KeyError(key)
		return found

	def get(self, key, default=None):
		try:
			v = self[key]
		except KeyError:
			return default
		else:
			return v

	def __len__(self):
		return len(self._items)


def main(args):
	with FastaReader(args.template) as fr:
		template = PrefixDict([])
		for record in fr:
			try:
				template.add(record.sequence.upper(), record.name)
			except ValueError:
				logger.error('Sequences in entry %r and %r are duplicate',
					record.name, template[record.sequence.upper()])
	logger.info('Read %d entries from template', len(template))

	with FastaReader(args.target) as fr:
		sequences = list(fr)

	# Rename
	renamed = 0
	for record in sequences:
		name = template.get(record.sequence.upper())
		if name is None:
			name = record.name + args.not_found
		else:
			renamed += 1
		record.name = name

	if args.sort:
		sequences = sorted(sequences, key=lambda s: natural_sort_key(s.name))
	for record in sequences:
		print('>{}\n{}'.format(record.name, record.sequence))
	logger.info('Wrote %s FASTA records (%d sequences found in template)', len(sequences), renamed)



"""
Run IgBLAST

This is a wrapper for the igblastn tool that has a somewhat simpler command-line
syntax and can also run IgBLAST in parallel. The result is printed to standard
output.
"""
import sys
import os
import multiprocessing
import subprocess
from itertools import islice
import pkg_resources

from sqt import SequenceReader
from sqt.utils import available_cpu_count


def add_arguments(parser):
	arg = parser.add_argument
	arg('--threads', '-t', '-j', type=int, default=available_cpu_count(),
		help='Number of threads. Default: no. of available CPUs (%(default)s)')
	arg('--penalty', type=int, choices=(-1, -2, -3, -4), default=None,
		help='BLAST mismatch penalty (default: -1)')
	arg('--species', default='rhesus_monkey',
		help='Which species (default: %(default)s)')
	arg('--limit', type=int, metavar='N',
		help='Limit processing to first N records')
	arg('database', help='Database directory. Must contain BLAST databases for '
		'SPECIES_V, SPECIES_D and SPECIES_J.')
	arg('fasta', help='File with original reads')


def run_igblast(fasta, database, species, penalty=None):
	"""
	fasta -- path to a FASTA file or data in FASTA format as a string
	database -- directory that contains IgBLAST databases. Files in that
	directory must be databases created by the makeblastdb program and have
	names organism_gene, such as "rhesus_monkey_V".

	Return IgBLAST’s output as a string.
	"""

	"""
	TODO
	Igblastwrapper has the databases in files named like this:
	$IGDATA/database/rhesus_monkey_IG_H_J
	Should we also include the _IG_H part?
	"""
	if fasta.startswith('>'):
		fasta_input = '-'
	else:
		fasta_input = fasta
	arguments = ['igblastn']
	for gene in 'V', 'D', 'J':
		arguments += ['-germline_db_{gene}'.format(gene=gene),
			os.path.join(database, '{species}_{gene}'.format(species=species, gene=gene))]
	if penalty is not None:
		arguments += ['-penalty', str(penalty)]
	# The empty aux suppresses a warning from IgBLAST. /dev/null does not work.
	empty_aux_path = pkg_resources.resource_filename('igdiscover', 'empty.aux')
	arguments += [
		'-auxiliary_data', empty_aux_path,
		'-organism', species,
		'-ig_seqtype', 'Ig',
		'-num_threads' , '1',
		'-domain_system', 'imgt',
		'-num_alignments_V', '1',
		'-num_alignments_D', '1',
		'-num_alignments_J', '1',
		'-outfmt', "7 sseqid qstart qseq sstart sseq pident slen evalue",
		'-query', fasta_input,
		'-out', '-',  # write to stdout
	]
	result = subprocess.check_output(arguments, input=fasta if fasta_input == '-' else None, universal_newlines=True)
	return result


def chunked_fasta(path, chunksize=500, limit=None):
	"""
	Split a FASTA file into chunks that contain at most chunksize records.
	Take only the first 'limit' records, or all if limit is None.
	"""
	def buf_to_fasta(b):
		return ''.join(">{}\n{}\n".format(r.name, r.sequence) for r in b)
	sr = SequenceReader(path)
	buf = []
	for record in islice(sr, 0, limit):
		buf.append(record)
		if len(buf) == chunksize:
			yield buf_to_fasta(buf)
			buf = []
	if len(buf) > 0:
		yield buf_to_fasta(buf)
	sr.close()


class Runner:
	"""nested functions cannot be pickled"""
	def __init__(self, dbpath, species, penalty):
		self.dbpath = dbpath
		self.species = species
		self.penalty = penalty

	def __call__(self, s):
		return run_igblast(s, self.dbpath, self.species, self.penalty)


def main(args):
	chunks = chunked_fasta(args.fasta, limit=args.limit)
	runner = Runner(args.database, args.species, args.penalty)
	with multiprocessing.Pool(args.threads) as pool:
		for result in pool.imap(runner, chunks, chunksize=1):
			sys.stdout.write(result)



"""
Count and plot V, D, J gene usage.
"""
import sys
import logging
import numpy as np
import pandas as pd
import matplotlib
matplotlib.use('pdf')
import matplotlib.pyplot as plt
from sqt import FastaReader

from .table import read_table
from .utils import natural_sort_key

logger = logging.getLogger(__name__)

def add_arguments(parser):
	arg = parser.add_argument
	arg('--gene', default='V', choices=('V', 'D', 'J'),
		help='Which type gene to count. Default: V')
	arg('--database', metavar='FASTA',
		help='FASTA file with reference gene sequences. The names are used to '
		'ensure all names appear in the plot')
	arg('table', help='Table with parsed and filtered IgBLAST results.')
	arg('plot', nargs='?', help='Plot file (png or pdf).')


def main(args):
	if args.database:
		with FastaReader(args.database) as fr:
			gene_names = [ record.name for record in fr ]
		gene_names.sort(key=natural_sort_key)
	else:
		gene_names = None
	table = read_table(args.table, log=True)

	columns = ('gene', 'count', 'unique_CDR3')
	if args.gene != 'V':
		columns += ('unique_V',)
	if args.gene != 'J':
		columns += ('unique_J',)
	column_name = '{}_gene'.format(args.gene)
	if len(table) > 0:
		rows = []
		groups = table.groupby(column_name)
		for gene, group in groups:
			count = len(group)
			unique_CDR3 = len(set(s for s in group.CDR3_nt if s))
			row = dict(gene=gene, count=count, unique_CDR3=unique_CDR3)
			if args.gene != 'V':
				row['unique_V'] = len(set(s for s in group.V_gene if s))
			if args.gene != 'J':
				row['unique_J'] = len(set(s for s in group.J_gene if s))
			rows.append(row)
		counts = pd.DataFrame(rows, columns=columns).set_index('gene')
	else:
		# Work around a pandas bug in reindex when the table is empty
		counts = pd.DataFrame(columns=columns, dtype=int).set_index('gene')

	# Make sure that always all gene names are listed even if no sequences
	# were assigned.
	if gene_names:
		counts = counts.reindex(gene_names, fill_value=0)

	counts.to_csv(sys.stdout, sep='\t')
	logger.info("Wrote CSV")

	if not args.plot:
		return

	fig = plt.figure(figsize=((50 + len(counts) * 5) / 25.4, 210/25.4))
	matplotlib.rcParams.update({'font.size': 14})
	ax = fig.gca()
	ax.set_title('{} gene usage'.format(args.gene))
	ax.set_xlabel('{} gene'.format(args.gene))
	ax.set_ylabel('Count')
	ax.set_xticks(np.arange(len(counts)) + 0.5)
	ax.set_xticklabels(counts.index, rotation='vertical')
	ax.grid(axis='x')
	ax.set_xlim((-0.25, len(counts)))
	ax.bar(np.arange(len(counts)), counts['count'])
	fig.set_tight_layout(True)
	fig.savefig(args.plot)
	logger.info("Wrote %s", args.plot)



"""
For each V gene, plot a clustermap of the sequences assigned to it.
"""
import os.path
import logging
import matplotlib
matplotlib.use('pdf')
import seaborn as sns
from .table import read_table
from .utils import downsampled, plural_s
from .cluster import cluster_sequences

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--minimum-group-size', '-m', metavar='N', type=int, default=200,
		help='Do not plot if there are less than N sequences for a gene. Default: %(default)s')
	arg('--gene', '-g', action='append', default=[],
		help='Compute consensus for this gene. Can be given multiple times. Default: Compute for all genes.')
	arg('--size', metavar='N', type=int, default=300,
		help='Show at most N sequences (with a matrix of size N x N). Default: %(default)s')
	arg('--ignore-J', action='store_true', default=False,
		help='Include also rows without J assignment or J%%SHM>0.')
	arg('--dpi', type=int, default=200, help='Resolution of output file. Default: %(default)s')
	arg('--no-title', dest='title', action='store_false', default=True,
		help='Do not add a title to the plot')
	arg('table', help='Table with parsed and filtered IgBLAST results')
	arg('directory', help='Save clustermaps as PNG into this directory', default=None)


def plot_clustermap(group, title, plotpath, size=300, dpi=200):
	"""
	Plot a clustermap for a specific V gene.

	size -- Downsample to this many sequences
	title

	Return the number of clusters.
	"""
	sequences = list(group.V_nt)
	sequences = downsampled(sequences, size)
	df, linkage, clusters = cluster_sequences(sequences)

	palette = sns.color_palette([(0.15, 0.15, 0.15)])
	palette += sns.color_palette('Spectral', n_colors=max(clusters), desat=0.9)
	row_colors = [ palette[cluster_id] for cluster_id in clusters ]
	cm = sns.clustermap(df,
			row_linkage=linkage,
			col_linkage=linkage,
			row_colors=row_colors,
			linewidths=None,
			linecolor='none',
			figsize=(210/25.4, 210/25.4),
			cmap='Blues',
			xticklabels=False,
			yticklabels=False
	)
	if title is not None:
		cm.fig.suptitle(title)
	cm.savefig(plotpath, dpi=dpi)

	# free the memory used by the plot
	import matplotlib.pyplot as plt
	plt.close('all')

	return len(set(clusters))


def main(args):
	if not os.path.exists(args.directory):
		os.mkdir(args.directory)
	table = read_table(args.table)

	# Discard rows with any mutation within J at all
	logger.info('%s rows read', len(table))
	if not args.ignore_J:
		# Discard rows with any mutation within J at all
		table = table[table.J_SHM == 0][:]
		logger.info('%s rows remain after discarding J%%SHM > 0', len(table))

	path_sanitizer = { ord(c): None for c in "\\/*?[]=\"'" }
	genes = frozenset(args.gene)
	n = 0
	too_few = 0
	for gene, group in table.groupby('V_gene'):
		if genes and gene not in genes:
			continue
		if len(group) < args.minimum_group_size:
			too_few += 1
			continue
		title = gene if args.title else None
		path = os.path.join(args.directory, gene.translate(path_sanitizer) + '.png')
		n_clusters = plot_clustermap(group, title, path, size=args.size, dpi=args.dpi)
		n += 1
		logger.info('Plotted %r with %d cluster%s', gene, n_clusters, plural_s(n_clusters))
		#for i, cons in enumerate(consensus_sequences):
			#print('>{}_cluster{}\n{}'.format(gene, ('red', 'blue')[i], cons))
			#print('number of Ns:', cons.count('N'))
		#if len(consensus_sequences) >= 2:
			#print('difference between consensuses:', edit_distance(*consensus_sequences[:2]))


	logger.info('%s plots created (%s skipped because of too few sequences)', n, too_few)



"""
Some utility functions that work on sequences and lists of sequences.
"""
import random
import hashlib
import os
from collections import OrderedDict
import re
import numpy as np
import yaml
from sqt.align import edit_distance, multialign, consensus
from sqt.dna import GENETIC_CODE


def downsampled(population, size):
	"""
	Return a random subsample of the population.

	Uses reservoir sampling. See https://en.wikipedia.org/wiki/Reservoir_sampling
	"""
	sample = population[:size]
	for index in range(size, len(population)):
		r = random.randint(0, index)
		if r < size:
			sample[r] = population[index]
	return sample


def distances(sequences, band=0.2):
	"""
	Compute all pairwise edit distances and return a square matrix.

	Entry [i,j] in the matrix is the edit distance between sequences[i]
	and sequences[j].
	"""
	m = np.zeros((len(sequences), len(sequences)), dtype=float)
	maxdiff = max((int(len(s) * band) for s in sequences), default=0)
	for i, s in enumerate(sequences):
		for j, t in enumerate(sequences):
			if i < j:
				m[j,i] = m[i,j] = min(maxdiff+1, edit_distance(s, t, maxdiff=maxdiff))
	return m


def iterative_consensus(sequences, program='muscle-medium', threshold=0.6,
		subsample_size=200, maximum_subsample_size=1600):
	"""
	Compute a consensus sequence of the given sequences, but do not use all
	sequences if there are many: First, try to compute the consensus from a
	small subsample. If there are 'N' bases, increase the subsample size and
	repeat until either there are no more 'N' bases, all available sequences
	have been used or maximum_subsample_size is reached.
	"""
	while True:
		sample = downsampled(sequences, subsample_size)
		aligned = multialign(OrderedDict(enumerate(sample)), program=program)
		cons = consensus(aligned, threshold=threshold).strip('N')
		if 'N' not in cons:
			# This consensus is good enough
			break
		if len(sequences) <= subsample_size:
			# We have already used all the sequences that are available
			break
		subsample_size *= 2
		if subsample_size > maximum_subsample_size:
			break
	return cons


def sequence_hash(s, digits=4):
	"""
	For a string, return a 'fingerprint' that looks like 'S1234' (the character
	'S' is fixed). The idea is that this allows one to quickly see whether two
	sequences are not identical.
	"""
	h = int(hashlib.md5(s.encode()).hexdigest()[-4:], base=16)
	return 'S' + str(h % 10**digits).rjust(digits, '0')


class SerialPool:
	"""
	An alternative to multiprocessing.Pool that runs things in serial for
	easier debugging
	"""
	def __init__(self, *args, **kwargs):
		pass

	def __enter__(self):
		return self

	def __exit__(self, *args):
		pass

	def imap(self, func, iterable, chunksize):
		for i in iterable:
			yield func(i)


def natural_sort_key(s, _nsre=re.compile('([0-9]+)')):
	"""
	Use this function as sorting key to sort in 'natural' order.

	>>> names = ['file10', 'file1.5', 'file1']
	>>> sorted(names, key=natural_sort_key)
	['file1', 'file1.5', 'file10']

	Source: http://stackoverflow.com/a/16090640/715090
	"""
	return [int(text) if text.isdigit() else text.lower()
		for text in re.split(_nsre, s)]


class UniqueNamer:
	"""
	Assign unique names by appending letters to already seen names.
	"""
	def __init__(self):
		self._names = set()

	def __call__(self, name):
		ext = 'A'
		new_name = name
		while new_name in self._names:
			if ext == '[':
				raise ValueError('Too many duplicate names')
			new_name = name + ext
			ext = chr(ord(ext) + 1)
		self._names.add(new_name)
		return new_name


class Merger:
	"""
	Merge similar items into one. To specify what "similar" means, implement
	the merged() method in a subclass.
	"""
	def __init__(self):
		self._items = []

	def add(self, item):
		# This method could possibly be made simpler if the graph structure
		# was made explicit.

		items = []
		m = None
		for existing_item in self._items:
			m = self.merged(existing_item, item)
			if m is None:
				items.append(existing_item)
			else:
				item = m
		items.append(item)
		self._items = items

	def __iter__(self):
		yield from sorted(self._items, key=lambda x: x.name)

	def merged(self, existing_item, item):
		"""
		If existing_item and item can be returned, this method must return
		a new item that represents a merged version of both. If they cannot
		be merged, it must return None.
		"""
		raise NotImplementedError("not implemented")


def relative_symlink(src, dst, force=False):
	"""
	Create a symbolic link in any directory.

	force -- if True, then overwrite an existing file/symlink
	"""
	if force:
		try:
			os.remove(dst)
		except FileNotFoundError:
			pass
	target = os.path.relpath(os.path.abspath(src), start=os.path.dirname(dst))
	os.symlink(target, dst)


class ConfigurationError(Exception):
	pass


class Config:
	DEFAULT_PATH = 'igdiscover.yaml'

	def __init__(self, file):
		# Set some defaults.
		self.species = 'human'
		self.merge_program = 'pear'
		self.flash_maximum_overlap = 300
		self.limit = None  # or an integer
		self.cluster_program = 'vsearch'
		self.multialign_program = 'muscle-fast'
		self.maximum_expected_errors = None  # or an integer
		self.minimum_merged_read_length = 300
		self.mismatch_penalty = None
		self.barcode_length = 0
		self.iterations = 3
		self.ignore_j = False
		self.subsample = 1000
		self.stranded = False
		self.forward_primers = None
		self.reverse_primers = None
		self.rename = True
		self.seed = 1
		self.exact_copies = None
		self.pre_germline_filter = dict(unique_cdr3s=2, unique_js=2, check_motifs=False,
			whitelist=True, cluster_size=0, differences=2, allow_stop=True)
		self.germline_filter = dict(unique_cdr3s=5, unique_js=3, check_motifs=False,
			whitelist=True, cluster_size=100, differences=2, allow_stop=False)
		self.library_name = os.path.basename(os.getcwd())

		self.read_from(file)

	def read_from(self, file):
		content = file.read()
		new_config = self.make_compatible(yaml.safe_load(content))
		for key in ('pre_germline_filter', 'germline_filter'):
			if key in new_config:
				self.__dict__[key].update(new_config[key])
				del new_config[key]
		self.__dict__.update(new_config)

	def make_compatible(self, config):
		"""
		Convert old-style configuration to new style. Raise ConfigurationError if configuration is invalid.
		Return updated config dict.
		"""
		if 'barcode_length' in config and ('barcode_length_5prime' in config or 'barcode_length_3prime' in config):
			raise ConfigurationError('Old-style configuration of barcode length via "barcode_length" option cannot be '
					'used at the same time as new-style configuration via "barcode_length_5prime" or "barcode_length_3prime"')
		if 'barcode_length_5prime' in config:
			config['barcode_length'] = config['barcode_length_5prime']
			del config['barcode_length_5prime']
		if 'barcode_length_3prime' in config:
			config['barcode_length'] = -config['barcode_length_3prime']
			del config['barcode_length_3prime']
		if 'seed' in config and config['seed'] is False:
			config['seed'] = None
		return config

	@classmethod
	def from_default_path(cls):
		with open(cls.DEFAULT_PATH) as f:
			return Config(file=f)


def nt_to_aa(s):
	"""Translate a nucleotide sequence to an amino acid sequence"""
	return ''.join(GENETIC_CODE.get(s[i:i+3], '*') for i in range(0, len(s), 3))


def has_stop(sequence):
	"""
	Return a boolean indicating whether the sequence has an internal stop codon.
	An incomplete codon at the end is allowed.

	>>> has_stop('GGG')
	False
	>>> has_stop('TAA')
	True
	>>> has_stop('GGGAC')
	False
	"""
	s = sequence[:len(sequence) // 3 * 3]
	return '*' in nt_to_aa(s)


def plural_s(n):
	return 's' if n != 1 else ''



"""
Function for reading the table created by the 'parse' subcommand.
"""
import os.path
import logging
import sqlite3
import pandas as pd
import numpy as np
from tempfile import TemporaryDirectory

logger = logging.getLogger(__name__)

_CACHING = False

# These columns contain string data
# convert them to str to avoid a PerformanceWarning
# TODO some of these are actually categorical or bool
_STRING_COLUMNS = [
	'V_gene',  # categorical
	'D_gene',  # categorical
	'J_gene',  # categorical
	'chain',  # categorical
	'stop',  # bool
	'productive',  # bool
	'UTR',
	'leader',
	'CDR1_nt',
	'CDR1_aa',
	'CDR2_nt',
	'CDR2_aa',
	'CDR3_nt',
	'CDR3_aa',
	'V_nt',
	'V_aa',
	'V_end',
	'VD_junction',
	'D_region',
	'DJ_junction',
	'J_nt',
	'VDJ_nt',
	'VDJ_aa',
	'name',
	'barcode',
	'race_G',
	'genomic_sequence',
]

_INTEGER_COLUMNS = ('V_errors', 'J_errors')

def _fix_columns(df):
	"""
	Changes DataFrame in-place
	"""
	# Convert all string columns to str to avoid a PerformanceWarning
	for col in _STRING_COLUMNS:
		if not col in df:
			continue
		df[col].fillna('', inplace=True)
		df[col] = df[col].astype('str')
		# Empty strings have been set to NaN by read_csv. Replacing
		# by the empty string avoids problems with groupby, which
		# ignores NaN values.
	# Columns that have any NaN values in them cannot be converted to
	# int due to a numpy limitation.
	for col in _INTEGER_COLUMNS:
		if all(df[col].notnull()):
			df[col] = df[col].astype(int)


def read_table(path, log=False):
	"""
	Read in the table created by the parse subcommand (typically named *.tab)
	"""
	if _CACHING:
		base, ext = os.path.splitext(path)
		h5path = base + '.h5'

		if not os.path.exists(h5path) or os.path.getmtime(path) > os.path.getmtime(h5path):
			logger.info('(Re-)Creating table cache %s', h5path)
			# Two or more processes may try to re-create the table at the same time.
			# We therefore write the table into a temporary file first and then move
			# it atomically.
			with TemporaryDirectory(dir=os.path.dirname(h5path)) as tempdir:
				temp_h5 = os.path.join(tempdir, 'db.h5')
				df = pd.read_csv(path, sep='\t') #true_values=['yes'], false_values=['no'])
				_fix_columns(df)
				df.to_hdf(temp_h5, 'table', complevel=3, complib='zlib')
				os.rename(temp_h5, h5path)
		d = pd.read_hdf(h5path, 'table')
	else:
		d = pd.read_csv(path, sep='\t')
		_fix_columns(d)
	if log:
		logger.info('%s rows in input table', len(d))
	return d



import pandas as pd
from scipy.spatial import distance
from scipy.cluster import hierarchy
from .utils import distances, iterative_consensus


def all_nodes(root):
	"""Return a list of all nodes of the tree, from left to right (iterative implementation)."""
	result = []
	path = [None]
	node = root
	while node is not None:
		if node.left is not None:
			path.append(node)
			node = node.left
		elif node.right is not None:
			result.append(node)
			node = node.right
		else:
			result.append(node)
			node = path.pop()
			if node is not None:
				result.append(node)
				node = node.right

	return result


def inner_nodes(root):
	"""
	Return a list of all inner nodes of the tree, from left to right.
	"""
	return [ node for node in all_nodes(root) if not node.is_leaf() ]


def collect_ids(root):
	"""
	Return a list of ids of all leaves of the given tree
	"""
	return [ node.id for node in all_nodes(root) if node.is_leaf() ]


def cluster_sequences(sequences, minsize=5):
	"""
	Cluster the given sequences into groups of similar sequences.

	Return a triple that contains a pandas.DataFrame with the edit distances,
	the linkage result, and a list that maps sequence ids to their cluster id.
	If an entry is zero in that list, it means that the sequence is not part of
	a cluster.
	"""
	matrix = distances(sequences)
	linkage = hierarchy.linkage(distance.squareform(matrix), method='average')
	# Linkage columns are:
	# 0, 1: merged clusters, 2: distance, 3: number of nodes in cluster
	inner = inner_nodes(hierarchy.to_tree(linkage))
	prev = linkage[:,2].max()  # highest distance
	clusters = [0] * len(sequences)
	cl = 1
	for n in inner:
		if prev/n.dist < 0.8 and n.left.count >= minsize and n.right.count >= minsize:
			for id in collect_ids(n.left):
				# Do not overwrite previously assigned ids
				if clusters[id] == 0:
					clusters[id] = cl
			cl += 1
		prev = n.dist

	return pd.DataFrame(matrix), linkage, clusters



from igdiscover.cluster import inner_nodes


def inner_nodes_recursive(root):
	"""
	Return a list of all inner nodes of the tree, from left to right.
	"""
	if root.is_leaf():
		return []
	return inner_nodes_recursive(root.left) + [root] + inner_nodes_recursive(root.right)


def collect_ids_recursive(root):
	"""
	Return a list of ids of all leaves of the given tree
	"""
	if root.is_leaf():
		return [root.id]
	return collect_ids_recursive(root.left) + collect_ids_recursive(root.right)


def test_inner_nodes():
	class Node:
		def __init__(self, value, left, right):
			self.value = value
			self.left = left
			self.right = right

		def is_leaf(self):
			return self.left is None and self.right is None

		def __repr__(self):
			# return 'Node({!r}, {!r}, {!r})'.format(self.value, self.left, self.right)
			return 'Node({!r}, ...)'.format(self.value)

	empty_tree = None
	assert inner_nodes(empty_tree) == []

	leaf = Node(0, None, None)
	assert inner_nodes(leaf) == []

	tree = Node(1, leaf, leaf)
	values = [ n.value for n in inner_nodes(tree) ]
	assert values == [1]

	tree = Node(1, Node(2, Node(3, leaf, leaf), leaf), leaf)
	values = [ n.value for n in inner_nodes(tree) ]
	assert values == [3, 2, 1]

	tree = Node(1, Node(2, Node(3, leaf, Node(4, leaf, leaf)), leaf), leaf)
	values = [ n.value for n in inner_nodes(tree) ]
	assert values == [3, 4, 2, 1]

	tree = Node(1,
		leaf,
		Node(2,
			leaf,
			Node(3,
	            Node(4,
	                leaf,
	                Node(5, leaf, leaf)
	            ),
	            Node(6, leaf, leaf)
			)
		)
	)
	values = [ n.value for n in inner_nodes(tree) ]
	assert values == [1, 2, 4, 5, 3, 6]



"""
Discover candidate new V genes within a single antibody library.

Existing V sequences are grouped by their V gene assignment, and within each
group, consensus sequences are computed.
"""
import csv
import hashlib
import logging
import sys
import os.path
from collections import OrderedDict, namedtuple, Counter
import multiprocessing
import random
from itertools import zip_longest
import numpy as np
import pandas as pd
from sqt import SequenceReader
from sqt.align import edit_distance
from sqt.utils import available_cpu_count

from .table import read_table
from .utils import iterative_consensus, sequence_hash, downsampled, SerialPool, Merger, has_stop
from .cluster import cluster_sequences
from .species import looks_like_V_gene

logger = logging.getLogger(__name__)


MINGROUPSIZE = 5
MAXIMUM_SUBSAMPLE_SIZE = 1600


def add_arguments(parser):
	arg = parser.add_argument
	arg('--threads', '-j', type=int, default=min(4, available_cpu_count()),
		help='Number of threads. Default: no. of available CPUs, but at most 4')
	arg('--seed', type=int, default=None,
		help='Seed value for random numbers for reproducible runs.')
	arg('--consensus-threshold', '-t', metavar='PERCENT', type=float, default=60,
		help='Threshold for consensus computation. Default: %(default)s%%')
	arg('--prefix', default='', metavar='PREFIX',
		help='Add PREFIX before sequence names')
	arg('--gene', '-g', action='append', default=[],
		help='Compute consensus for this gene. Can be given multiple times. '
			'Default: Compute for all genes.')
	arg('--left', '-l', type=float, metavar='ERROR-RATE',
		help='For consensus, include only sequences that have at least this '
			'error rate (in percent). Default: %(default)s', default=0)
	arg('--right', '-r', type=float, metavar='ERROR-RATE',
		help='For consensus, include only sequences that have at most this '
			'error rate (in percent). Default: %(default)s', default=100)
	arg('--window-width', '-w', type=float, metavar='PERCENT',
		help='Compute consensus for all PERCENT-wide windows. Set to 0 to '
			'disable. Default: %(default)s', default=2)
	arg('--no-cluster', dest='cluster', action='store_false', default=True,
		help='Do not run linkage cluster analysis.')
	arg('--max-n-bases', type=int, default=0, metavar='MAXN',
		help='Remove rows that have more than MAXN "N" nucleotides. If >0, an '
			'N_bases column is added. Default: %(default)s')
	arg('--subsample', metavar='N', type=int, default=1000,
		help='When clustering, use N randomly chosen sequences. Default: %(default)s')
	arg('--table-output', '-o', metavar='DIRECTORY',
		help='Output tables for all analyzed genes to DIRECTORY. '
			'Files will be named <GENE>.tab.')
	arg('--database', metavar='FASTA', default=None,
		help='FASTA file with V genes. If provided, differences between consensus '
			'and database will be computed.')
	arg('--ignore-J', action='store_true', default=False,
		help='Include also rows without J assignment or J%%SHM>0.')
	arg('--approx', action='store_true', default=False,
		help='Count also approximate matches (adds three columns to output table).')
	arg('--error-rate', metavar='PERCENT', type=float, default=1,
		help='When finding approximate V gene matches, allow PERCENT errors. '
			'Default: %(default)s')
	arg('--exact-copies', metavar='N', type=int, default=1,
		help='When subsampling, first pick rows whose V gene sequences'
			'have at least N exact copies in the input. Default: %(default)s')
	arg('table', help='Table with parsed IgBLAST results')


Groupinfo = namedtuple('Groupinfo', 'count unique_J unique_CDR3')

SiblingInfo = namedtuple('SiblingInfo', 'sequence requested name group')


class SiblingMerger(Merger):
	"""
	Merge very similar consensus sequences into single entries. This could be
	seen as a type of clustering using very specific criteria. Two sequences
	are merged if one is the prefix of the other, allowing differences where
	one of the sequences has an 'N' base.
	"""
	def merged(self, s, t):
		chars = []
		for c1, c2 in zip_longest(s.sequence, t.sequence):
			if c1 is None:
				c = c2
			elif c2 is None:
				c = c1
			elif c1 == 'N':
				c = c2
			elif c2 == 'N':
				c = c1
			elif c1 != c2:
				return None
			else:
				assert c1 == c2
				c = c1
			chars.append(c)
		seq = ''.join(chars)
		requested = s.requested or t.requested
		name = s.name + ';' + t.name
		# take union of groups
		group = pd.concat([s.group, t.group]).groupby(level=0).last()
		return SiblingInfo(seq, requested, name, group)


class Discoverer:
	"""
	Discover candidates for novel V genes.
	"""
	def __init__(self, database, windows, left, right, cluster, table_output,
			prefix, consensus_threshold, v_error_rate, downsample,
			cluster_subsample_size, approx_columns, max_n_bases, exact_copies,
			seed):
		self.database = database
		self.windows = windows
		self.left = left
		self.right = right
		self.cluster = cluster
		self.table_output = table_output
		self.prefix = prefix
		self.consensus_threshold = consensus_threshold
		self.v_error_rate = v_error_rate
		self.downsample = downsample
		self.cluster_subsample_size = cluster_subsample_size
		self.approx_columns = approx_columns
		self.max_n_bases = max_n_bases
		self.exact_copies = exact_copies
		self.seed = seed

	def _sibling_sequence(self, gene, group):
		"""
		For a given group, compute a consensus sequence over the V gene sequences
		in that group.

		If the found sibling is slightly longer or shorter than the version in
		the database, adjust it so it corresponds to the database version exactly.
		"""
		sequence = iterative_consensus(list(group.V_nt), program='muscle-medium',
			threshold=self.consensus_threshold/100,
			maximum_subsample_size=self.downsample)
		if gene in self.database:
			database_sequence = self.database[gene]
			if sequence.startswith(database_sequence) or database_sequence.startswith(sequence):
				return database_sequence
		return sequence

	def _guess_chain(self, group):
		"""
		Return a guess for the chain type of a given group
		"""
		return Counter(group.chain).most_common()[0][0]

	def _collect_siblings(self, gene, group):
		"""
		gene -- gene name
		group -- pandas.DataFrame of sequences assigned to that gene
		"""
		group = group.copy()
		database_sequence = self.database.get(gene, None)
		database_sequence_found = False
		for left, right in self.windows:
			left, right = float(left), float(right)
			group_in_window = group[(left <= group.V_SHM) & (group.V_SHM < right)]
			if len(group_in_window) < MINGROUPSIZE:
				continue
			sibling = self._sibling_sequence(gene, group_in_window)
			if left == int(left):
				left = int(left)
			if right == int(right):
				right = int(right)
			requested = (left, right) == (self.left, self.right)
			name = '{}-{}'.format(left, right)
			if sibling == database_sequence:
				database_sequence_found = True
			yield SiblingInfo(sibling, requested, name, group_in_window)

		if self.cluster:
			if self.exact_copies > 1:
				# Preferentially pick those sequences (for subsampling) that have
				# multiple exact copies, then fill up with the others
				exact_group = group[group.copies >= self.exact_copies]
				indices = downsampled(list(exact_group.index), self.cluster_subsample_size)
				if len(indices) < self.cluster_subsample_size:
					not_exact_group = group[group.copies < self.exact_copies]
					indices.extend(downsampled(list(not_exact_group.index),
						self.cluster_subsample_size - len(indices)))
			else:
				indices = downsampled(list(group.index), self.cluster_subsample_size)
			sequences = list(group.V_nt.loc[indices])
			df, linkage, clusters = cluster_sequences(sequences)
			logger.info('Clustering %d sequences (downsampled to %d) assigned to %r gave %d cluster(s)',
				len(group), len(indices), gene, len(set(clusters)))
			cluster_indices = [ [] for _ in range(max(clusters) + 1) ]
			for i, cluster_id in enumerate(clusters):
				cluster_indices[cluster_id].append(indices[i])

			cl = 1
			for ind in cluster_indices:
				group_in_window = group.loc[ind]
				if len(group_in_window) < MINGROUPSIZE:
					logger.info('Skipping cluster %d because it is too small', ind)
					continue
				sibling = self._sibling_sequence(gene, group_in_window)
				name = 'cl{}'.format(cl)
				if sibling == database_sequence:
					database_sequence_found = True
				yield SiblingInfo(sibling, False, name, group_in_window)
				cl += 1

		if database_sequence:
			# Hard-coded threshold of 1% V SHM
			group_in_window = group[group.V_SHM <= 1.0]
			if len(group_in_window) >= MINGROUPSIZE:
				if not database_sequence_found:
					logger.info('Database sequence for %r is expressed, but missing from candidates. '
						'Re-adding it.', gene)
				yield SiblingInfo(database_sequence, False, 'db', group_in_window)

	def set_random_seed(self, name):
		"""Set random seed depending on gene name and seed given to constructor"""
		h = hashlib.md5(name.encode()).digest()[:4]
		n = int.from_bytes(h, byteorder='big')
		random.seed(n + self.seed)

	def __call__(self, args):
		"""
		Discover new V genes. args is a tuple (gene, group)
		gene -- name of the gene
		group -- a pandas DataFrame with the group corresponding to the gene
		"""
		gene, group = args
		self.set_random_seed(gene)
		siblings = SiblingMerger()
		for sibling in self._collect_siblings(gene, group):
			siblings.add(sibling)

		candidates = []
		for sibling_info in siblings:
			sibling = sibling_info.sequence
			n_bases = sibling.count('N')
			if n_bases > self.max_n_bases:
				logger.debug('Sibling %s has too many N bases', sibling_info.name)
				continue
			prefix_identical = group.V_nt.map(lambda s: s.startswith(sibling) or sibling.startswith(s))
			group_exact_V = group[prefix_identical]
			if self.approx_columns:
				group['consensus_diff'] = [ edit_distance(v_nt, sibling) for v_nt in group.V_nt ]
				group_approximate_V = group[group.consensus_diff <= len(sibling) * self.v_error_rate]

			groups = (
				('window', sibling_info.group),
				('exact', group_exact_V))
			if self.approx_columns:
				groups += (('approx', group_approximate_V), )
			info = dict()
			for key, g in groups:
				unique_J = len(set(s for s in g.J_gene if s))
				unique_CDR3 = len(set(s for s in g.CDR3_nt if s))
				count = len(g.index)
				info[key] = Groupinfo(count=count, unique_J=unique_J, unique_CDR3=unique_CDR3)
			if gene in self.database:
				database_diff = edit_distance(sibling, self.database[gene])
			else:
				database_diff = None

			# Build the Candidate
			# TODO use UniqueNamer here
			sequence_id = '{}{}_{}'.format(self.prefix, gene.rsplit('_S', 1)[0], sequence_hash(sibling))

			if self.approx_columns:
				assert info['exact'].count <= info['approx'].count
				approx = info['approx'].count
				Js_approx = info['approx'].unique_J
				CDR3s_approx = info['approx'].unique_CDR3
			else:
				approx = None
				Js_approx = None
				CDR3s_approx = None

			chain = self._guess_chain(sibling_info.group)
			try:
				ratio = info['exact'].count / info['exact'].unique_CDR3
			except ZeroDivisionError:
				ratio = 0
			candidate = Candidate(
				name=sequence_id,
				source=gene,
				chain=chain,
				cluster=sibling_info.name if len(sibling_info.group) < len(group) else 'all',
				cluster_size=info['window'].count,
				Js=info['window'].unique_J,
				CDR3s=info['window'].unique_CDR3,
				exact=info['exact'].count,
				Js_exact=info['exact'].unique_J,
				CDR3s_exact=info['exact'].unique_CDR3,
				CDR3_exact_ratio='{:.2f}'.format(ratio),
				approx=approx,
				Js_approx=Js_approx,
				CDR3s_approx=CDR3s_approx,
				N_bases=n_bases,
				database_diff=database_diff,
				has_stop=int(has_stop(sibling)),
				looks_like_V=int(looks_like_V_gene(sibling, chain)),
				consensus=sibling,
			)
			candidates.append(candidate)

			# If a window was requested via --left/--right, write the 'approx'
			# subset to a separate file.
			if self.table_output and self.approx_columns and any(si.requested for si in sibling_info) and len(group_approximate_V) > 0:
				if not os.path.exists(self.table_output):
					os.mkdir(self.table_output)
				path = os.path.join(self.table_output, gene + '.tab')
				group_approximate_V.sort('consensus_diff').to_csv(path, sep='\t')
				logger.info('Wrote %s for window %s-%s', path, self.left, self.right)
		return candidates


Candidate = namedtuple('Candidate', [
	'name',
	'source',
	'chain',
	'cluster',
	'cluster_size',
	'Js',
	'CDR3s',
	'exact',
	'Js_exact',
	'CDR3s_exact',
	'CDR3_exact_ratio',
	'approx',
	'Js_approx',
	'CDR3s_approx',
	'N_bases',
	'database_diff',
	'has_stop',
	'looks_like_V',
	'consensus'
])


def count_prefixes(sequences):
	"""
	Count how often each sequence occurs in the given list of
	sequences. If one sequence is the prefix of another one,
	they are considered to be 'identical'.

	Return a dictionary that maps sequence to count.

	>>> r = count_prefixes(['A', 'BB', 'CD', 'CDE', 'CDEF'])
	>>> r == {'A': 1, 'BB': 1, 'CD': 3, 'CDE': 3, 'CDEF': 3}
	True
	"""
	sequences = sorted(sequences)
	sequences.append('GUARD')
	prev = 'X'
	start = 0
	count = dict()
	for i, s in enumerate(sequences):
		if not s.startswith(prev):
			# end of a run
			for j in range(start, i):
				count[sequences[j]] = i - start
			start = i
		prev = s
	return count


def main(args):
	v_error_rate = args.error_rate / 100
	assert 0 <= v_error_rate <= 1

	if args.database:
		with SequenceReader(args.database) as sr:
			database = { record.name: record.sequence.upper() for record in sr }
	else:
		database = dict()

	if args.seed:
		seed = args.seed
	else:
		seed = random.randrange(10**6)
	logger.info('Use --seed=%d to reproduce this run', seed)

	table = read_table(args.table)
	table = table.loc[:,('name', 'chain', 'V_gene', 'J_gene', 'V_nt', 'CDR3_nt', 'V_SHM', 'J_SHM')].copy()

	logger.info('%s rows read', len(table))
	if not args.ignore_J:
		# Discard rows with any mutation within J at all
		table = table[table.J_SHM == 0][:]
		logger.info('%s rows remain after discarding J%%SHM > 0', len(table))

	if args.approx:
		logger.info('Approximate comparisons between V gene sequence and consensus '
			'allow %.1f%% errors.', v_error_rate*100)

	if args.exact_copies > 1:
		multiplicities = count_prefixes(table.V_nt)
		table['copies'] = table.V_nt.map(multiplicities)
		logger.info('%s rows contain V sequences with at least %s copies',
			sum(table.copies >= args.exact_copies), args.exact_copies)

	columns = list(Candidate._fields)
	if not args.approx:
		for col in ['approx', 'Js_approx', 'CDR3s_approx']:
			columns.remove(col)
	if not args.max_n_bases:
		columns.remove('N_bases')
	writer = csv.DictWriter(sys.stdout, fieldnames=columns, delimiter='\t', extrasaction='ignore')
	writer.writeheader()
	genes = set(args.gene)
	if args.window_width:
		windows = [ (start, start + args.window_width) for start in np.arange(0, 20, args.window_width) ]
		logger.info('Using an error rate window of %.1f%% to %.1f%%', args.left, args.right)
		windows.append((args.left, args.right))
	else:
		windows = []

	groups = []
	for gene, group in table.groupby('V_gene'):
		if genes and gene not in genes:
			continue
		if len(group) < MINGROUPSIZE:
			continue
		groups.append((gene, group))

	discoverer = Discoverer(database, windows, args.left, args.right, args.cluster,
		args.table_output, args.prefix, args.consensus_threshold, v_error_rate,
		MAXIMUM_SUBSAMPLE_SIZE, cluster_subsample_size=args.subsample,
		approx_columns=args.approx, max_n_bases=args.max_n_bases, exact_copies=args.exact_copies,
		seed=seed)
	n_consensus = 0

	Pool = SerialPool if args.threads == 1 else multiprocessing.Pool
	with Pool(args.threads) as pool:
		for candidates in pool.imap(discoverer, groups, chunksize=1):
			for candidate in candidates:
				writer.writerow(candidate._asdict())
			sys.stdout.flush()
			n_consensus += len(candidates)
	logger.info('%s consensus sequences for %s gene(s) computed', n_consensus, len(groups))



"""
Filter V gene candidates (germline and pre-germline filter)

After candidates for novel V genes have been found with the 'discover'
subcommand, this script is used to filter the candidates and make sure that
only true germline genes remain ("germline filter" and "pre-germline filter").
The following filtering and processing steps are performed:

* Discard sequences with N bases
* Discard sequences that come from a consensus over too few source sequences 
* Discard sequences with too few unique CDR3s (CDR3s_exact column)
* Discard sequences with too few unique Js (Js_exact column)
* Discard sequences identical to one of the database sequences (if DB given)
* Discard sequences that do not match a set of known good motifs
* Discard sequences that contain a stop codon (has_stop column)
* Discard near-duplicate sequences

If you provide a whitelist of sequences, then the candidates that appear on it
* are not checked for the cluster size criterion,
* do not need to match a set of known good motifs,
* are never considered near-duplicates,
* are allowed to contain a stop codon.

The filtered table is written to standard output.
"""
import logging
from collections import namedtuple
import pandas as pd
from sqt import FastaReader
from sqt.align import edit_distance

from .utils import UniqueNamer, Merger

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--cluster-size', type=int, metavar='N', default=0,
		help='Consensus must represent at least N sequences. '
		'Default: %(default)s')
	arg('--max-differences', type=int, metavar='MAXDIFF', default=1,
		help='Merge sequences if they have at most MAXDIFF differences. '
		'The one with more CDR3s is kept. Default: %(default)s')
	arg('--minimum-db-diff', '-b', type=int, metavar='N', default=0,
		help='Sequences must have at least N differences to the database '
		'sequence. Default: %(default)s')
	arg('--maximum-N', '-N', type=int, metavar='COUNT', default=0,
		help='Sequences must have at most COUNT "N" bases. Default: %(default)s')
	arg('--unique-CDR3', '--CDR3s', type=int, metavar='N', default=1,
		help='Sequences must have at least N unique CDR3s within exact sequence matches. '
		'Default: %(default)s')
	# The default for unique-J is 0 because we might work on data without
	# any assigned J genes.
	arg('--unique-J', type=int, metavar='N', default=0,
		help='Sequences must have at least N unique Js within exact sequence matches. '
		'Default: %(default)s')
	arg('--looks-like-V', action='store_true', default=False,
		help='Sequences must look like V genes (uses the looks_like_V column). '
		'Default: Column is ignored')
	arg('--allow-stop', action='store_true', default=False,
		help='Allow stop codons in sequences (uses the has_stop column).'
			'Default: Do not allow stop codons.')
	arg('--whitelist', metavar='FASTA', default=[], action='append',
		help='Sequences that are never discarded or merged with others, '
			'even if criteria for discarding them would apply.')
	arg('--fasta', metavar='FILE', help='Write new database in FASTA format to FILE')
	arg('tables', metavar='CANDIDATES.TAB',
		help='Tables (one or more) created by the "discover" command',
		nargs='+')


SequenceInfo = namedtuple('SequenceInfo', 'sequence name CDR3s_exact whitelisted row')


class SequenceMerger(Merger):
	"""
	Merge sequences that are sufficiently similar into single entries.
	"""
	def __init__(self, max_differences):
		super().__init__()
		self._max_differences = max_differences

	def merged(self, s, t):
		"""
		Given two SequenceInfo objects, decide whether to discard one of them and which one.
		This is for merging similar sequences.

		Two sequences are considered to be similar if their edit distance is at most
		max_differences (see constructor). If one of the sequences is longer, the 'overhanging'
		bases are ignored at either the 5' end or the 3' end, whichever gives the lower
		edit distance.

		Sequences that are whitelisted are never discarded.
		If two sequences are similar and none of them is whitelisted, then the one with the
		higher number of unique CDR3s is kept.

		Return None if both objects should be kept.
		Return the object to keep otherwise.
		"""
		if len(s.sequence) > len(t.sequence):
			s, t = t, s  # make s always the shorter sequence
		s_seq = s.sequence
		if len(s_seq) != len(t.sequence):
			t_prefix = t.sequence[:len(s_seq)]
			t_suffix = t.sequence[-len(s_seq):]
			dist_prefix = edit_distance(s_seq, t_prefix, self._max_differences)
			dist_suffix = edit_distance(s_seq, t_suffix, self._max_differences)
			dist = min(dist_prefix, dist_suffix)
		else:
			dist = edit_distance(s_seq, t.sequence, self._max_differences)

		if dist > self._max_differences:
			return None  # keep both
		if s.whitelisted and t.whitelisted:
			return None  # keep both
		if s.whitelisted:
			return s
		if t.whitelisted:
			return t
		# No sequence is whitelisted if we arrive here
		if len(s_seq) < len(t.sequence):
			return t
		if s.CDR3s_exact >= t.CDR3s_exact:
			return s
		return t


def main(args):
	merger = SequenceMerger(args.max_differences)

	whitelist = dict()
	for path in args.whitelist:
		for record in FastaReader(path):
			whitelist[record.sequence.upper()] = record.name
	logger.info('%d unique sequences in whitelist', len(whitelist))

	def whitelist_dist(sequence):
		if sequence in whitelist:
			return 0, whitelist[sequence]
		mindist = len(sequence)
		distances = []
		for seq, name in whitelist.items():
			ed = edit_distance(seq, sequence, maxdiff=mindist)
			distances.append((ed, name))
			if ed == 1:
				# We know ed does not get smaller because the
				# 'sequence in whitelist' check
				# above covers that
				return ed, name
			mindist = min(mindist, ed)
		distance, name = min(distances)
		return distance, name

	# Read in tables
	total_unfiltered = 0
	overall_table = None
	for path in args.tables:
		table = pd.read_csv(path, sep='\t')
		# TODO remove this after deprecation period
		table.rename(columns=dict(
			consensus_seqs='cluster_size',
			window_seqs='cluster_size',
			subset_seqs='cluster_size'), inplace=True)

		i = list(table.columns).index('consensus')
		# whitelist_diff distinguishes between 0 and >0 only
		# at this point. Accurate edit distances are computed later.
		whitelist_diff = [ (0 if s in whitelist else -1) for s in table['consensus'] ]
		table.insert(i, 'whitelist_diff', pd.Series(whitelist_diff, index=table.index, dtype=int))
		table.insert(i+1, 'closest_whitelist', pd.Series('', index=table.index))

		unfiltered_length = len(table)
		table = table[table.database_diff >= args.minimum_db_diff]
		if 'N_bases' in table.columns:
			table = table[table.N_bases <= args.maximum_N]
		table = table[table.CDR3s_exact >= args.unique_CDR3]
		table = table[table.Js_exact >= args.unique_J]
		if args.looks_like_V:
			table = table[(table.looks_like_V == 1) | (table.whitelist_diff == 0)]
		if not args.allow_stop:
			table = table[(table.has_stop == 0) | (table.whitelist_diff == 0)]
		table = table[(table.cluster_size >= args.cluster_size) | (table.whitelist_diff == 0)]
		table = table.dropna()
		logger.info('Table read from %r contains %s candidate V gene sequences. '
			'%s remain after filtering', path,
			unfiltered_length, len(table))
		if args.whitelist:
			logger.info('Of those, %d are protected by the whitelist', sum(table.whitelist_diff == 0))
		total_unfiltered += unfiltered_length

		if overall_table is None:
			overall_table = table
		else:
			overall_table.append(table)
	del table
	if len(args.tables) > 1:
		logger.info('Read %s tables with %s entries total. '
			'After filtering, %s entries remain.', len(args.tables),
			total_unfiltered, len(overall_table))

	for _, row in overall_table.iterrows():
		merger.add(SequenceInfo(row['consensus'], row['name'], row['CDR3s_exact'],
			row['whitelist_diff'] == 0, row.name))  # row.name is the index of the row. It is not row['name'].

	# Discard near-duplicates
	overall_table['is_duplicate'] = pd.Series(True, index=overall_table.index, dtype=bool)
	for info in merger:
		overall_table.loc[info.row, 'is_duplicate'] = False
	overall_table = overall_table[~overall_table.is_duplicate].copy()
	del overall_table['is_duplicate']

	# Name sequences
	overall_table['name'] = overall_table['name'].apply(UniqueNamer())
	overall_table.sort_values(['name'], inplace=True)

	# Because whitelist_dist() is expensive, this is run when
	# all of the filtering has already been done
	if whitelist:
		for row in overall_table.itertuples():
			distance, name = whitelist_dist(overall_table.loc[row[0], 'consensus'])
			overall_table.loc[row[0], 'closest_whitelist'] = name
			overall_table.loc[row[0], 'whitelist_diff'] = distance
	else:
		overall_table.whitelist_diff.replace(-1, '', inplace=True)
	print(overall_table.to_csv(sep='\t', index=False, float_format='%.1f'), end='')

	if args.fasta:
		with open(args.fasta, 'w') as f:
			for _, row in overall_table.iterrows():
				print('>{}\n{}'.format(row['name'], row['consensus']), file=f)

	logger.info('%d sequences in new database', len(overall_table))
